#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "math.h"
#include "assert.h"

#include <FL/Fl.H>
#include <FL/Fl_Button.H>
#include <FL/fl_file_chooser.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Select_Browser.H>
#include <FL/Fl_Check_Browser.H>
#include <FL/Fl_Select_Browser.H>
#include <FL/Fl_Shared_Image.H>
#include <FL/Fl_Value_Input.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Multiline_Input.H>

#include "List.h"
#include "cwindow.h"
#include "twindow.h"
#include "owindow.h"
// #include "lwindow.h"

#include "GMapTile.h"
#include "GMap.h"


extern bool quit;
extern GMapsCWindow *cw;
extern GMapsTWindow *tw;
extern GMapsOWindow *ow;
// extern GMapsLWindow *lw;
extern int mode;
extern GMap *map;
extern List<GMapTile> tiles;
extern int n_objects;
extern char *object_name[];
extern int object_parameters[];
extern char *object_parameter_name[27][3];
extern char last_file_name[256];

int grid_size=16;

Fl_Double_Window *properties_dlg=0;

/* Button Callbacks: */ 

void gm_quit_cb(Fl_Widget *, void *);
void gm_load_cb(Fl_Widget *, void *);
void gm_save_cb(Fl_Widget *, void *);
void gm_clear_cb(Fl_Widget *, void *);
void gm_tiles_cb(Fl_Widget *, void *);
void gm_objects_cb(Fl_Widget *, void *);
void gm_clear_layer_cb(Fl_Widget *, void *);


GMapsCWindow::GMapsCWindow(int x,int y) : Fl_Double_Window(x,y,120+512+1+150+120,320+1+130,"GMaps") 
{
	Fl_Button *b;
	b=new Fl_Button(5,5,110,20,"QUIT");
	b->callback(gm_quit_cb,0);
	b=new Fl_Button(5,30,110,20,"LOAD");
	b->callback(gm_load_cb,0);
	b=new Fl_Button(5,55,110,20,"SAVE");
	b->callback(gm_save_cb,0);
	b=new Fl_Button(5,100,110,20,"CLEAR");
	b->callback(gm_clear_cb,0);
	b=new Fl_Button(5,125,110,20,"TILES");
	b->callback(gm_tiles_cb,0);
	b=new Fl_Button(5,150,110,20,"OBJECTS");
	b->callback(gm_objects_cb,0);

	m_visible_list=new Fl_Check_Browser(5+120+512,50,140,120);

	b=new Fl_Button(5+120+512+150,50,110,18,"Clear Layer 1");
	b->callback(gm_clear_layer_cb,0);
	b=new Fl_Button(5+120+512+150,70,110,18,"Clear Layer 2");
	b->callback(gm_clear_layer_cb,(void *)1);
	b=new Fl_Button(5+120+512+150,90,110,18,"Clear Layer 3");
	b->callback(gm_clear_layer_cb,(void *)2);
	b=new Fl_Button(5+120+512+150,110,110,18,"Clear Layer 4");
	b->callback(gm_clear_layer_cb,(void *)3);
	b=new Fl_Button(5+120+512+150,130,110,18,"Clear Layer 5");
	b->callback(gm_clear_layer_cb,(void *)4);

	m_script_editor=new Fl_Multiline_Input(5,325,120+512+150+110,120);

	end();

	callback(gm_quit_cb);

	update();
} /* GMapsCWindow::GMapsCWindow */ 

GMapsCWindow::~GMapsCWindow()
{
	quit=true;
} /* GMapsCWindow::~GMapsCWindow */ 


void gm_load_cb(Fl_Widget *,void *)
{
	FILE *fp;
	char *filename;

	filename=fl_file_chooser("Load map","*.gmp","*.gmp");

	if (filename!=0) {
		fp=fopen(filename,"r");
		if (fp!=0) {
			if (map!=0) delete map;
			map=new GMap(fp,&tiles);
			cw->m_script_editor->value(map->get_scripts());
			cw->update();
			fclose(fp);

			{
				int i,j=0;
				for(i=0;filename[i]!=0;i++) if (filename[i]=='/' || filename[i]=='\\') j=i+1;
				strcpy(last_file_name,filename+j);
			}
		} /* if */ 
	} /* if */ 
} /* gm_load_cb */ 


void gm_save_cb(Fl_Widget *,void *)
{
	FILE *fp;
	char *filename;
	char fn[256]="*.gmp";

	if (last_file_name[0]!=0) {
		strcpy(fn,last_file_name);
	} // if 

	filename=fl_file_chooser("Save map","*.gmp",fn);

	if (filename!=0) {
		fp=fopen(filename,"w");
		if (fp!=0) {
			map->set_scripts((char *)cw->m_script_editor->value());
			map->save(fp);
			fclose(fp);
		
			{
				int i,j=0;
				for(i=0;filename[i]!=0;i++) if (filename[i]=='/' || filename[i]=='\\') j=i+1;
				strcpy(last_file_name,filename+j);
			}
		} /* if */ 
	} /* if */ 
} /* gm_save_cb */ 


void gm_quit_cb(Fl_Widget *, void *)
{
	quit=true;
} /* gm_quit_cb */ 


void gm_clear_cb(Fl_Widget *, void *)
{
	if (map!=0) delete map;
	map=new GMap(32*16,20*16);
	cw->update();
} /* gm_clear_cb */ 


void gm_clear_layer_cb(Fl_Widget *, void *l)
{
	int layer=(int)l;

	map->clear_layer(layer);
	cw->update();
} /* gm_clear_cb */ 



void gm_tiles_cb(Fl_Widget *, void *)
{
	if (ow!=0) ow->hide();

	if (tw==0) tw=new GMapsTWindow(5,250);
	tw->show();

	mode=1;
} /* gm_tiles_cb */ 


void gm_objects_cb(Fl_Widget *, void *)
{
	if (tw!=0) tw->hide();

	if (ow==0) ow=new GMapsOWindow(5,250);
	ow->show();

	mode=2;
} /* gm_objects_cb */ 




void map_coordinates(int mx,int my,int *x,int *y)
{
	*x=mx-(mx%grid_size);
	*y=my-(my%grid_size);
} /* map_coordinates */ 



void GMapsCWindow::draw()
{
	damage(FL_DAMAGE_ALL);

	Fl_Double_Window::draw();
	fl_rectf(120,0,512+1,320+1,0,0,0);

	if (last_file_name[0]!=0) {
		fl_draw(last_file_name,5+120+512,15);
	} // if 


	{
		char tmp[80]="X,Y: -,-";
		if (this==Fl::belowmouse()) {
			int rx,ry;
			map_coordinates((int)Fl::event_x()-120,(int)Fl::event_y(),&rx,&ry);
			if (rx>=0 && ry>=0 &&
				rx<512 && ry<320) {
				sprintf(tmp,"X,Y: %i,%i",rx/grid_size,ry/grid_size);
			} // if 
		} // if
		fl_draw(tmp,10,210);
	}

	fl_draw("Visible Layers:",5+120+512,40);
	
	fl_draw("Help:",25+120+512,210);

	fl_draw("Add: left click",25+120+512,240);
	fl_draw("Delete: right click",25+120+512,260);
	fl_draw("Properties: CTRL + left click",25+120+512,280);

	// Draw the map:

	if (map!=0) {
		// draw map:
		{
			int i;
			for(i=0;i<map->get_n_layers();i++) {
				if (get_v_list()->checked(i+1)) map->draw(120,0,i);
			} // for  
		}
	} // if 




	// draw grid:
	{
		int i,j;
		fl_color(FL_WHITE);
		for(i=0;i<=map->get_dx();i+=grid_size) {
			fl_line(120+i,0,120+i,map->get_dy());
		} // for  
		for(j=0;j<=map->get_dy();j+=grid_size) {
			fl_line(120,j,120+map->get_dx(),j);
		} // for  
	}


	if (map!=0) {
		// Draw the pointer:
		switch(mode) {
		case 0:
				break;
		case 1:
				if (this==Fl::belowmouse()) {
					int rx,ry;
					char *text_sel=(tw!=0 ? tw->get_selected():0);
					int sel=-1;

					{
						int i=1;
						GMapTile *t;
						tiles.Rewind();
						while(sel==-1 && tiles.Iterate(t)) {
							if (text_sel!=0 && strcmp(t->get_name(),text_sel)==0) sel=i;
							i++;
						} // while
					}
					if (sel>0) {
						map_coordinates((int)Fl::event_x()-120,(int)Fl::event_y(),&rx,&ry);

						if (tw!=0 && sel>0 && sel<(tiles.Length()+1) && rx>=0) {
							tiles[sel-1]->get_sfc()->draw(120+rx,ry);
						} // if  
					} // if 
				} // if
				break;
		case 2: 
				if (this==Fl::belowmouse()) {				
					int sel=(ow!=0 ? ow->get_selected():0);
					int rx,ry;
					map_coordinates((int)Fl::event_x()-120,(int)Fl::event_y(),&rx,&ry);


					if (ow!=0 && sel>0 && sel<(n_objects+1) && rx>=0) {
						GMap::object_tile(object_name[sel-1],0)->get_sfc()->draw(rx+120,ry);
					} // if  

				}
				break;
		} // switch  

	} // if


} /* GMapsCWindow::draw */ 



void dlg_delete_cb(Fl_Widget *, void *ptr)
{
	GObjectPlace *op=(GObjectPlace *)ptr;

	if (map!=0) map->delete_object(op);

	if (properties_dlg!=0) {
		delete properties_dlg;
		properties_dlg=0;
	} /* if */  
} /* gm_addlayer_cb */ 



void dlg_cangevalue_cb(Fl_Widget *w, void *ptr)
{
	int *v=(int *)ptr;

	*v=(int)(((Fl_Value_Input *)w)->value());

} /* gm_addlayer_cb */ 



/*

enum Fl_Event {	// events
  FL_NO_EVENT		= 0,
  FL_PUSH		= 1,
  FL_RELEASE		= 2,
  FL_ENTER		= 3,
  FL_LEAVE		= 4,
  FL_DRAG		= 5,
  FL_FOCUS		= 6,
  FL_UNFOCUS		= 7,
  FL_KEYDOWN		= 8,
  FL_KEYUP		= 9,
  FL_CLOSE		= 10,
  FL_MOVE		= 11,
  FL_SHORTCUT		= 12,
  FL_DEACTIVATE		= 13,
  FL_ACTIVATE		= 14,
  FL_HIDE		= 15,
  FL_SHOW		= 16,
  FL_PASTE		= 17,
  FL_SELECTIONCLEAR	= 18,
  FL_MOUSEWHEEL		= 19,
  FL_DND_ENTER		= 20,
  FL_DND_DRAG		= 21,
  FL_DND_LEAVE		= 22,
  FL_DND_RELEASE	= 23
};

 */






int GMapsCWindow::handle(int event)
{
	redraw();

	if (event==FL_MOVE) {
		Fl::belowmouse(this);
		return 1;
	} // if 

//	printf("Event: %i (%i)\n",event,visible_r());
//	if (event==FL_HIDE && visible_focus()) quit=true;
//	if (event==FL_CLOSE || event==FL_HIDE) quit=true;

	if (event==FL_PUSH) {
		int x=Fl::event_x()-120;
		int y=Fl::event_y();	

		// LEFT BUTTON:
		if (mode==1 && x>=0 && x<512 && y<320 && Fl::event_button()==1 && (Fl::event_state()&FL_COMMAND)==0) {
			if (tw!=0) {
				char *text_sel=(tw!=0 ? tw->get_selected():0);
				int x,y;
				int t_sel=-1;

				{
					int i=1;
					GMapTile *t;
					tiles.Rewind();
					while(t_sel==-1 && tiles.Iterate(t)) {
						if (strcmp(t->get_name(),text_sel)==0) t_sel=i;
						i++;
					} // while
					assert(t_sel>0);
				}

				map_coordinates((int)Fl::event_x()-120,(int)Fl::event_y(),&x,&y);
				if (t_sel>0 && t_sel<(tiles.Length()+1)) {
//					int layer=get_e_list()->value()-1;
					int layer=0;

					if (tiles[t_sel-1]->get_name()[0]=='b' &&
						tiles[t_sel-1]->get_name()[1]=='g' &&
						tiles[t_sel-1]->get_name()[2]=='_') layer=0;
					if (tiles[t_sel-1]->get_name()[0]=='g' &&
						tiles[t_sel-1]->get_name()[1]=='r' &&
						tiles[t_sel-1]->get_name()[2]=='_') layer=1;
/*
					if (strcmp(tiles[t_sel-1]->get_name(),"gr_tile-ground-l")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tile-ground-m")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tile-ground-r")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tileblue-ground-l")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tileblue-ground-m")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tileblue-ground-r")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tilegreen-ground-l")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tilegreen-ground-m")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tilegreen-ground-r")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tile-bridge-l")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tile-bridge-m")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tile-bridge-r")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tileblue-bridge-l")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tileblue-bridge-m")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tileblue-bridge-r")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tilegreen-bridge-l")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tilegreen-bridge-m")==0 ||
						strcmp(tiles[t_sel-1]->get_name(),"gr_tilegreen-bridge-r")==0) layer=2;
*/
					if (tiles[t_sel-1]->get_name()[0]=='w' &&
						tiles[t_sel-1]->get_name()[1]=='l' &&
						tiles[t_sel-1]->get_name()[2]=='_') layer=3;

					map->add_tile(layer,x,y,tiles[t_sel-1]);
				} // if 
			} // if
			return 1;
		} /* if */ 

		if (mode==2 && x>=0 && x<512 && y<320 && Fl::event_button()==1 && (Fl::event_state()&FL_COMMAND)==0) {
			if (ow!=0) {
				int o_sel=(ow!=0 ? ow->get_selected():0);
				int x,y;
				map_coordinates((int)Fl::event_x()-120,(int)Fl::event_y(),&x,&y);
				// objects go to layer 2 ("Layer 3" in the interface)
				if (o_sel>0 && o_sel<(n_objects+1)) {
					if (strcmp(object_name[o_sel-1],"GO_rope")==0 ||
						strcmp(object_name[o_sel-1],"GO_ropetop")==0 ||
						strcmp(object_name[o_sel-1],"GO_ropebottom")==0) {
						map->add_object(3,x,y,object_name[o_sel-1]);
					} else {
						map->add_object(2,x,y,object_name[o_sel-1]);
					} // if 
				} // if 
//				map->add_tile(get_e_list()->value()-1,x,y,tiles[t_sel-1]);
			} // if
			return 1;
		} /* if */  

		// CTRL + LEFT BUTTON:
		if (Fl::event_button()==1 && (Fl::event_state()&FL_COMMAND)!=0 && x>=0 && x<512 && y<320) {
			GObjectPlace *op;
			int i;

			for(i=map->get_n_layers()-1;i>=0;i--) {
				op=map->get_object(i,(int)Fl::event_x()-120,(int)Fl::event_y());

				if (op!=0) {
					// Create context dialogue:
					{
						Fl_Button *b;
						Fl_Value_Input *v;
						if (properties_dlg!=0) delete properties_dlg;

						properties_dlg=new Fl_Double_Window((int)Fl::event_x(),(int)Fl::event_y(),160,256,"Properties");
						b=new Fl_Button(5,5,150,20,"DELETE");
						b->callback(dlg_delete_cb,(void *)op);

						// Properties:
						{
							int n=-1;
							int j;

							for(j=0;j<n_objects;j++) if (strcmp(object_name[j],op->m_name)==0) n=j;
							if (n==-1) return 1;

							while(op->m_parameters.Length()<object_parameters[n]) op->m_parameters.Add(new int(0));
							for(j=0;j<object_parameters[n];j++) {								
								new Fl_Box(5,35+j*50,100,15,object_parameter_name[n][j]);
								v=new Fl_Value_Input(5,55+j*50,100,20,object_parameter_name[n][j]);
								v->value(*(op->m_parameters[j]));
								v->callback(dlg_cangevalue_cb,(void *)op->m_parameters[j]);
							} /* for */ 
						}
						properties_dlg->end();

						properties_dlg->show();
					}
					return 1;
				} /* if */ 
			} // for  
			return 1;
		} 

		// RIGHT BUTTON:
		if (Fl::event_button()==3 && x>=0 && x<512 && y<320) {
			GObjectPlace *op;
			int i;

			for(i=map->get_n_layers()-1;i>=0;i--) {
				op=map->get_object(i,(int)Fl::event_x()-120,(int)Fl::event_y());

				if (op!=0) {
					map->delete_object(op);
					return 1;
				} else {
					if (get_v_list()->checked(i+1)) if (map->delete_tile(i,(int)Fl::event_x()-120,(int)Fl::event_y())) return 1;					
				} /* if */ 
			} // for  
			return 1;
		} 
	} /* if */ 

	return Fl_Double_Window::handle(event);
} /* GMapsCWindow::handle */ 


void GMapsCWindow::update(void)
{
	int i;
	char tmp[256];

	m_visible_list->clear();
//	m_editing_list->clear();

	for(i=0;i<map->get_n_layers();i++) {
		sprintf(tmp,"Layer %i",i+1);
		m_visible_list->add(tmp);
//		m_editing_list->add(tmp);
//		m_editing_list->select(i+1,0);
	} /* for */ 

	m_visible_list->check_all();
//	m_editing_list->select(1);
} /* GMapsCWindow::update */ 

